#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : BaseOOF
# @Time         : 2019-06-23 20:07
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
from abc import abstractmethod

import time
import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold, RepeatedStratifiedKFold, KFold, cross_val_predict, cross_validate

from lightgbm.sklearn import LGBMClassifier


class BaseOOF(object):
    """cross_val_predict"""

    def __init__(self, estimator=LGBMClassifier(), cv=5, random_state=None, n_repeats=None):
        self.estimator = estimator
        if n_repeats:
            self._kf = RepeatedStratifiedKFold(cv, True, random_state)
            self._num_preds = cv * n_repeats
        else:
            self._kf = StratifiedKFold(cv, True, random_state)
            self._num_preds = cv

    @abstractmethod
    def _fit(self, eval_set):
        """入参eval_set = [(X_train, y_train), (X_valid, y_valid)]"""
        raise NotImplementedError

    @abstractmethod
    def _predict(self, X):
        """"""
        raise NotImplementedError

    def fit(self, X, y, X_test, feval=None, mean_type='arithmetic'):
        """输入数组
        :param X:
        :param y:
        :param X_test:
        :param feval:
        :param ensemble: https://www.kaggle.com/c/severstal-steel-defect-detection/discussion/107716
            harmonic mean ≤ geometric mean ≤ arithmetic mean ≤ square mean
            调和平均数      ≤ 几何平均数       ≤ 算术平均数        ≤ 平方平均数
        :return:
        """

        self.oof_train = np.zeros(len(X))
        self.oof_test = np.zeros((len(X_test), self._num_preds))
        for n_fold, (train_index, valid_index) in enumerate(self._kf.split(X, y)):
            print("\n\033[94mFold %s started at %s\033[0m" % (n_fold + 1, time.ctime()))
            X_train, y_train = X[train_index], y[train_index]
            X_valid, y_valid = X[valid_index], y[valid_index]
            eval_set = [(X_train, y_train), (X_valid, y_valid)]

            ########################################################################
            # TODO: 多分类
            # 重写fit/preidct
            self.estimator = self._fit(eval_set)

            self.oof_train[valid_index] = self.estimator.predict_proba(X_valid)[:, 1]
            self.oof_test[:, n_fold] = self.estimator.predict_proba(X_test)[:, 1]
            ########################################################################

        # 输出需要的结果
        self.oof_test_rank = pd.DataFrame(self.oof_test).rank().mean(1) / len(self.oof_test)
        self.oof_test = self.oof_test.mean(1)
        if feval:
            if hasattr(feval, '__repr__'):
                metric_name = feval.__repr__().split()[1][:32].title()
            else:
                metric_name = "Score"
            score = feval(y, self.oof_train)
            print("\n\033[94mCV %s: %s ended at %s\033[0m" % (metric_name, score, time.ctime()))

    def oof_save(self, file=None):
        if file is None:
            file = self.estimator.__str__().split('(')[0][:32]
            file = '%s👍%s.csv' % (file, time.ctime())
        assert isinstance(file, str)
        _ = np.append(self.oof_train, self.oof_test)
        pd.DataFrame(_, columns='oof: train+test') \
            .to_csv(file, index=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : efs
# @Time         : 2019-07-26 13:49
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

"""
枚举：2^n
"""

from sklearn.neighbors import KNeighborsClassifier
from sklearn.datasets import load_iris
from mlxtend.feature_selection import ExhaustiveFeatureSelector as EFS

iris = load_iris()
X = iris.data
y = iris.target

knn = KNeighborsClassifier(n_neighbors=3)

efs1 = EFS(knn,
           min_features=1,
           max_features=4,
           scoring='accuracy',
           print_progress=True,
           cv=5)

efs1 = efs1.fit(X, y)

print('Best accuracy score: %.2f' % efs1.best_score_)
print('Best subset (indices):', efs1.best_idx_)
print('Best subset (corresponding names):', efs1.best_feature_names_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : sfs
# @Time         : 2019-07-26 13:41
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
"""
efs的优化版：根据 scoring 筛选特征
顺序特征选择算法的贪婪搜索算法家族，用于减少初始d维特征空间到ķ维特征空间，其中ķ<d 。
特征选择算法背后的动机是自动选择与问题最相关的特征子集。
特征选择的目标是双重的：我们希望通过去除不相关的特征或噪声来提高计算效率并减少模型的泛化误差。
如果嵌入式特征选择（例如，像LASSO这样的正则化惩罚）不适用，则诸如顺序特征选择之类的包装器方法尤其有用。
"""


from mlxtend.feature_selection import SequentialFeatureSelector as SFS
from mlxtend.plotting import plot_sequential_feature_selection as plot_sfs
from sklearn.neighbors import NearestNeighbors
from sklearn.feature_selection import GenericUnivariateSelect, \
    SelectPercentile, SelectKBest, f_classif, mutual_info_classif, RFE

from lightgbm import LGBMClassifier
import matplotlib.pyplot as plt

X = ...
y = ...

sfs = SFS(LGBMClassifier(),
          k_features=10,
          forward=True,
          floating=False,
          verbose=2,
          scoring='roc_auc',
          cv=3,
          n_jobs=-1)

sfs.fit(X, y)

#  plot
fig1 = plot_sfs(sfs.get_metric_dict(), kind='std_dev')

plt.ylim([0.8, 1])
plt.title('Sequential Forward Selection (w. StdDev)')
plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : OOFlgb
# @Time         : 2019-06-23 23:41
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
from lightgbm import LGBMClassifier

from .BaseOOF import BaseOOF
from sklearn.metrics import roc_auc_score


class LGB(BaseOOF):
    def __init__(self, estimator=LGBMClassifier(), cv=5, random_state=None, **kwargs):
        super().__init__(**kwargs)

    def _fit(self, eval_set):
        _ = self.estimator.fit(
            *eval_set[0],
            eval_set=eval_set,
            eval_metric="auc",
            early_stopping_rounds=100,
            verbose=100)
        return _


if __name__ == "__main__":
    from sklearn.datasets import make_classification

    X, y = make_classification()
    clf = LGB()
    clf.fit(X, y, X, feval=roc_auc_score)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'demo'
__author__ = 'JieYuan'
__mtime__ = '19-3-19'
"""
from lightgbm import LGBMClassifier
from sklearn.datasets import make_classification
from yuan.models.bayes_opt import Optimizer, ParamsBounds
from pprint import pprint

X, y = make_classification(
    n_samples=1000,
    n_features=45,
    n_informative=12,
    n_redundant=7,
    random_state=134985745,
)

estimator = LGBMClassifier(random_state=123456, subsample_freq=5, n_estimators=256)
pbounds = {
    'learning_rate': (0.001, 0.1),
    'num_leaves': (2 ** 2, 2 ** 8),
    'min_split_gain': (0.001, 1),
    'min_child_weight': (0.001, 10),
    'min_child_samples': (16, 128),
    'subsample': (0.6, 1),
    'colsample_bytree': (0.6, 1),
    'reg_alpha': (0.01, 10),
    'reg_lambda': (0.01, 10)
}

opt = Optimizer(estimator, pbounds)
pprint(opt.pbounds)
pprint(opt.scaled_pbounds)

opt.fit(X, y, 30)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'example'
__author__ = 'JieYuan'
__mtime__ = '19-2-18'
"""
import numpy as np
from hyperopt import fmin, hp, tpe, Trials, STATUS_OK
from hyperopt.pyll import scope
import lightgbm as lgb
from sklearn.datasets import load_iris

N_FOLDS = 5
bayes_trials = Trials()
X, y = load_iris(True)
# Create the dataset
train_set = lgb.Dataset(X[:100], y[:100])


def objective(params, n_folds=N_FOLDS):
    """Objective function for Gradient Boosting Machine Hyperparameter Tuning"""

    # Perform n_fold cross validation with hyperparameters
    # Use early stopping and evalute based on ROC AUC
    params['verbosity'] = -1
    cv_results = lgb.cv(params, train_set, nfold=n_folds, num_boost_round=10000, early_stopping_rounds=100,
                        metrics='auc', seed=50, verbose_eval=50)
    # 此部分为核心代码，

    # Extract the best score
    best_score = max(cv_results['auc-mean'])
    # Loss must be minimized
    loss = 1 - best_score

    # Dictionary with information for evaluation
    return {'loss': loss, 'params': params, 'status': STATUS_OK}


# Define the search space
space = {
    'num_leaves': scope.int(hp.quniform('num_leaves', 2 ** 4, 2 ** 8, 1)),
    'learning_rate': hp.loguniform('learning_rate', np.log(1e-5), np.log(1)),

    'min_child_weight': hp.uniform('min_child_weight', 0, 0.01),
    'min_child_samples': scope.int(hp.quniform('min_child_samples', 8, 32, 2)),

    # 采样
    'subsample': hp.uniform('subsample', 0.5, 1),
    'subsample_freq': scope.int(hp.quniform('subsample_freq', 1, 5, 1)),
    'colsample_bytree': hp.uniform('colsample_by_tree', 0.5, 1.0),
    # 正则
    'min_split_gain': hp.uniform('min_split_gain', 0, 1),
    'reg_alpha': hp.loguniform('reg_alpha', np.log(1e-5), np.log(1e5)),
    'reg_lambda': hp.loguniform('reg_lambda', np.log(1e-5), np.log(1e5)),
}

MAX_EVALS = 100

# Optimize
best = fmin(fn=objective, space=space, algo=tpe.suggest,
            max_evals=MAX_EVALS, trials=bayes_trials, verbose=1)


print(bayes_trials)



# xgb
# space = {
#     'n_estimators': hp.quniform('n_estimators', 100, 1000, 1),
#     'eta': hp.quniform('eta', 0.025, 0.5, 0.025),
#     'max_depth': hp.quniform('max_depth', 1, 13, 1),
#     'min_child_weight': hp.quniform('min_child_weight', 1, 6, 1),
#     'subsample': hp.quniform('subsample', 0.5, 1, 0.05),
#     'gamma': hp.quniform('gamma', 0.5, 1, 0.05),
#     'colsample_bytree': hp.quniform('colsample_bytree', 0.5, 1, 0.05),
#     'num_class': 9,
#     'eval_metric': 'mlogloss',
#     'objective': 'multi:softprob',
#     'nthread': 6,
#     'silent': 1
# }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : oof
# @Time         : 2019-09-07 23:44
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from tql.algo_ml.oof import OOF, LGB
from sklearn.datasets import make_classification
from sklearn.metrics import roc_auc_score
X, y = make_classification(1000)
clf = LGB()
clf.fit(X, y, X, feval=roc_auc_score)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : DataFlods
# @Time         : 2019-06-23 22:29
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
from abc import abstractmethod
from sklearn.model_selection import StratifiedKFold, RepeatedStratifiedKFold, KFold


class DataFolds(object):

    def __init__(self, X, y, cv=5):
        self.X = X
        self.y = y
        self.cv = 5

    def __getitem__(self, index):
        """Gets batch at position `index`.

        Arguments:
            index: position of the batch in the Sequence.

        Returns:
            A batch
        """
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        """Number of batch in the Sequence.

        Returns:
            The number of batches in the Sequence.
        """

        raise NotImplementedError

    def on_epoch_end(self):
        """Method called at the end of every epoch.
        """
        pass

    def __iter__(self):
        """Create a generator that iterate over the Sequence."""
        for item in (self[i] for i in range(len(self))):
            yield item
